/*
  spamdyke -- a filter for stopping spam at connection time.
  Copyright (C) 2007 Sam Clippinger (samc (at) silence (dot) org)

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License version 2 as
  published by the Free Software Foundation.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
#include <stdio.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <sys/select.h>
#include <errno.h>
#include <string.h>
#include <sys/stat.h>
#include <syslog.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <arpa/nameser.h>
#include <arpa/nameser_compat.h>
#include <resolv.h>
#include <ctype.h>
#include <time.h>
#include <stdlib.h>
#include <netdb.h>

#define DEFAULT_PATH                    "/bin:/usr/bin:/usr/local/bin:/var/qmail/bin:/home/vpopmail/bin"
#define DEFAULT_TIMEOUT_COMMAND_SECS    60
#define DEFAULT_TIMEOUT_CONNECTION_SECS 0

#define ENVIRONMENT_TLS                 "SMTPS"
#define STRLEN_ENVIRONMENT_TLS          5
#define ENVIRONMENT_REMOTE_IP           "TCPREMOTEIP"
#define STRLEN_ENVIRONMENT_REMOTE_IP    11
#define ENVIRONMENT_REMOTE_NAME         "TCPREMOTEHOST"
#define STRLEN_ENVIRONMENT_REMOTE_NAME  13
#define ENVIRONMENT_PATH                "PATH"
#define STRLEN_ENVIRONMENT_PATH         4

#define STDIN_FD                        0
#define STDOUT_FD                       1

#define MKDIR_MODE                      0700

#define MAX_ADDRESS                     511
#define MAX_PATH                        4095
#define MAX_BUF                         4095
#define MAX_POLICY                      100
#define MINVAL(a,b)                     (((a) > (b)) ? (b) : (a))
#define MAXVAL(a,b)                     (((a) < (b)) ? (b) : (a))

#define FILTER_INTERCEPT                0
#define FILTER_PASS                     1
#define FILTER_QUIT                     2

#define LOCALHOST_IP                    "127.0.0.1"
#define STRLEN_LOCALHOST_IP             9
#define LOCALHOST_NAME                  "localhost"
#define STRLEN_LOCALHOST_NAME           9

#define ALPHABET_FILENAME               "abcdefghijklmnopqrstuvwxyz0123456789@_-.ABCDEFGHIJKLMNOPQRSTUVWXYZ"
#define REPLACEMENT_FILENAME            '_'

#define REJECT_SEVERITY_NONE            "250 "
#define REJECT_SEVERITY_HELO            "220 "
#define REJECT_SEVERITY_QUIT            "221 "
#define REJECT_SEVERITY_TEMPORARY       "451 "
#define REJECT_SEVERITY_PERMANENT       "500 "
#define STRLEN_REJECT_SEVERITY          4
#define REJECT_CRLF                     "\r\n"
#define STRLEN_REJECT_CRLF              2

#define SMTP_HELO_1                     "HELO"
#define STRLEN_SMTP_HELO_1              4
#define SMTP_HELO_2                     "EHLO"
#define STRLEN_SMTP_HELO_2              4
#define SMTP_MAIL_FROM                  "MAIL FROM"
#define STRLEN_SMTP_MAIL_FROM           9
#define SMTP_RCPT_TO                    "RCPT TO"
#define STRLEN_SMTP_RCPT_TO             7
#define SMTP_DATA                       "DATA"
#define STRLEN_SMTP_DATA                4
#define SMTP_DATA_END                   "."
#define STRLEN_SMTP_DATA_END            1
#define SMTP_QUIT                       "QUIT"
#define STRLEN_SMTP_QUIT                4
#define SMTP_RSET                       "RSET"
#define STRLEN_SMTP_RSET                4

#define CHILD_QUIT                      "QUIT\r\n"
#define STRLEN_CHILD_QUIT               6

#define SENDER_ADDRESS_NONE             "_none"

#define SHORT_SUCCESS                   "ALLOWED"

#define ERROR_URL                       " See: "
#define STRLEN_ERROR_URL                6

#define ERROR_RCPT_TO                   "Too many recipients. Try the remaining addresses again later."
#define STRLEN_ERROR_RCPT_TO            61
#define SHORT_ERROR_RCPT_TO             "DENIED_TOO_MANY_RECIPIENTS"

#define ERROR_RCPT_TO_LOCAL             "Improper recipient address. Try supplying a domain name."
#define STRLEN_ERROR_RCPT_TO_LOCAL      56
#define SHORT_ERROR_RCPT_TO_LOCAL       "DENIED_UNQUALIFIED_RECIPIENT"

#define ERROR_GRAYLISTED                "Your address has been graylisted. Try again later."
#define STRLEN_ERROR_GRAYLISTED         50
#define SHORT_ERROR_GRAYLISTED          "DENIED_GRAYLISTED"

#define ERROR_RDNS_MISSING              "Refused. You have no reverse DNS entry."
#define STRLEN_ERROR_RDNS_MISSING       39
#define SHORT_ERROR_RDNS_MISSING        "DENIED_RDNS_MISSING"

#define ERROR_RDNS_RESOLVE              "Refused. Your reverse DNS entry does not resolve."
#define STRLEN_ERROR_RDNS_RESOLVE       49
#define SHORT_ERROR_RDNS_RESOLVE        "DENIED_RDNS_RESOLVE"

#define ERROR_IP_IN_NAME_CC             "Refused. Your reverse DNS entry contains your IP address and a country code."
#define STRLEN_ERROR_IP_IN_NAME_CC      76
#define SHORT_ERROR_IP_IN_NAME_CC       "DENIED_IP_IN_CC_RDNS"

#define ERROR_IP_IN_NAME                "Refused. Your reverse DNS entry contains your IP address and a banned keyword."
#define STRLEN_ERROR_IP_IN_NAME         78
#define SHORT_ERROR_IP_IN_NAME          "DENIED_IP_IN_RDNS"

#define ERROR_EARLYTALKER               "Refused. You are not following the SMTP protocol."
#define STRLEN_ERROR_EARLYTALKER        49
#define SHORT_ERROR_EARLYTALKER         "DENIED_EARLYTALKER"

#define ERROR_BLACKLIST_NAME            "Refused. Your domain name is blacklisted."
#define STRLEN_ERROR_BLACKLIST_NAME     41
#define SHORT_ERROR_BLACKLIST_NAME      "DENIED_BLACKLIST_NAME"

#define ERROR_BLACKLIST_IP              "Refused. Your IP address is blacklisted."
#define STRLEN_ERROR_BLACKLIST_IP       40
#define SHORT_ERROR_BLACKLIST_IP        "DENIED_BLACKLIST_IP"

#define ERROR_TIMEOUT                   "Timeout. Talk faster next time."
#define STRLEN_ERROR_TIMEOUT            31
#define SHORT_ERROR_TIMEOUT             "TIMEOUT"

#define ERROR_SENDER_BLACKLISTED        "Refused. Your sender address has been blacklisted."
#define STRLEN_ERROR_SENDER_BLACKLISTED 50
#define SHORT_ERROR_SENDER_BLACKLISTED  "DENIED_SENDER_BLACKLISTED"

#define SHORT_ERROR_RBL                 "DENIED_RBL_MATCH"

#define SYSLOG_IDENTIFIER               "spamdyke"
#define SYSLOG_MISSING_DATA             "(unknown)"
#define STRLEN_SYSLOG_MISSING_DATA      9
#define SYSLOG_ERROR_GRAYLIST_FILE      "ERROR: cannot write to graylist file "
#define SYSLOG_ERROR_RCPTHOSTS_INIT     "ERROR: cannot initialize rcpthosts facility -- incorrect working directory?"
#define SYSLOG_ERROR_RCPTHOSTS          "ERROR: cannot use rcpthosts facility -- incorrect working directory?"
#define SYSLOG_ERROR_MOVE_DESCRIPTORS   "ERROR: unable to move file descriptors: "
#define SYSLOG_ERROR_FORK               "ERROR: unable to fork: "
#define SYSLOG_ERROR_PIPE               "ERROR: unable to create pipe: "
#define SYSLOG_ERROR_EXEC               "ERROR: unable to execute child process "
#define SYSLOG_ERROR_NO_MEMORY          "ERROR: unable to allocate memory for string copy."
#define SYSLOG_ERROR_OPEN_KEYWORDS      "ERROR: unable to open keywords file: "
#define SYSLOG_ERROR_OPEN_SEARCH        "ERROR: unable to open file for searching: "
#define SYSLOG_ERROR_RES_INIT           "ERROR: unable to initialize resolver library."
#define SYSLOG_ERROR_DN_EXPAND          "ERROR: dn_expand failed; this could indicate a problem with the nameserver."
#define SYSLOG_ERROR_TYPE_TXT           "ERROR: dns query did not include a TXT record as requested; this could indicate a problem with the name server."
#define SYSLOG_ERROR_DNS_RESPONSE       "ERROR: bad or invalid dns response; this could indicate a problem with the name server."
#define SYSLOG_ERROR_STAT               "ERROR: unable to stat graylist folder "
#define SYSLOG_ERROR_MKDIR              "ERROR: unable to create graylist folder "
#define SYSLOG_ERROR_OPEN_LOG           "ERROR: unable to open traffic log file: "

#define LOG_ACTION_NONE                         -1
#define LOG_ACTION_REMOTE_FROM                  0
#define LOG_ACTION_CHILD_FROM                   1
#define LOG_ACTION_CHILD_FROM_DISCARDED         2
#define LOG_ACTION_FILTER_FROM                  3
#define LOG_ACTION_FILTER_TO                    4
#define LOG_ACTION_PREFIX                       (char *[]){ ">>> ", "<<< ", "<XX ", "<FF ", "FF>" }

struct filter_settings
  {
  char *server_name;
  int strlen_server_name;
  char *server_ip;
  int strlen_server_ip;

  char *graylist_dir;
  long graylist_min_secs;
  long graylist_max_secs;
  int max_rcpt_to;
  int use_syslog;
  char *policy_location;
  int strlen_policy_location;
  char *log_dir;
  char *blacklist_sender_path;
  char *local_domains_path;

  int timeout_connection;
  int timeout_command;

  int inside_data;
  int num_rcpt_to;
  int local_sender;

  char sender_address[MAX_ADDRESS + 1];
  char recipient_address[MAX_ADDRESS + 1];
  char graylist_path[MAX_PATH + 1];

  char *reject_message;
  int strlen_reject_message;
  char *transient_reject_message;
  char *short_reject_message;
  char reject_message_buf[MAX_BUF + 1];
  };

void init_settings(struct filter_settings *target_settings)
  {
  target_settings->server_name = NULL;
  target_settings->strlen_server_name = 0;
  target_settings->server_ip = NULL;
  target_settings->strlen_server_ip = 0;

  target_settings->graylist_dir = NULL;
  target_settings->graylist_min_secs = 0;
  target_settings->graylist_max_secs = 0;
  target_settings->max_rcpt_to = 0;
  target_settings->use_syslog = 0;
  target_settings->policy_location = NULL;
  target_settings->strlen_policy_location = 0;
  target_settings->log_dir = NULL;
  target_settings->blacklist_sender_path = NULL;
  target_settings->local_domains_path = NULL;

  target_settings->timeout_connection = DEFAULT_TIMEOUT_CONNECTION_SECS;
  target_settings->timeout_command = DEFAULT_TIMEOUT_COMMAND_SECS;

  target_settings->inside_data = 0;
  target_settings->num_rcpt_to = 0;
  target_settings->local_sender = 1;

  target_settings->sender_address[0] = '\0';
  target_settings->recipient_address[0] = '\0';
  target_settings->graylist_path[0] = '\0';

  target_settings->reject_message = NULL;
  target_settings->strlen_reject_message = 0;
  target_settings->transient_reject_message = NULL;
  target_settings->short_reject_message = NULL;
  target_settings->reject_message_buf[0] = '\0';

  return;
  }

void usage()
  {
  printf(
    "USAGE: spamdyke [ -clrR ] [ -d LOCAL_DOMAINS_FILE [ -a MAX_RCPTS ] ]\n"
    "         [ -g GRAYLIST_DIR [-m MIN_SECS] [-M MAX_SECS] ]\n"
    "         [ -e DELAY_SECS ] [ -k RDNS_KEYWORD_FILE ]\n"
    "         [ -b RDNS_BLACKLIST_FILE ] [ -B IP_BLACKLIST_FILE ]\n"
    "         [ -w RDNS_WHITELIST_FILE ] [ -W IP_WHITELIST_FILE ]\n"
    "         [ -u POLICY_LOCATION ] [ -x RBL ] [ -L LOG_DIR ]\n"
    "         [ -s SENDER_BLACKLIST_FILE ]\n"
    "         [ -t MAX_CONNECTION_SECONDS ] [ -T MAX_COMMAND_SECONDS ]\n"
    "         path_to_exec [exec_args]\n"
    "\n"
    "-a Allow a maximum of MAX_RCPTS recipients per connection for non-local\n"
    "   senders. Requires -d. Default: unlimited recipients per connection.\n"
    "\n"
    "-b Search RDNS_BLACKLIST_FILE file for the base domain name of the remote\n"
    "   server's rDNS name. If it is found, reject the connection.\n"
    "\n"
    "-B Search IP_BLACKLIST_FILE file for the IP address of the remote server. If\n"
    "   it is found, reject the connection.\n"
    "\n"
    "-c Check the remote server's rDNS name for its IP address. If it is found AND\n"
    "   the rDNS name ends in a country code, reject the connection. Default: do not\n"
    "   check for a country code.\n"
    "\n"
    "-d Use LOCAL_DOMAINS_FILE file as a list of locally hosted domains (used by\n"
    "   other features).\n"
    "\n"
    "-e Delay the greeting banner DELAY_SECS to see if the remote server starts\n"
    "   sending data early.  If it does, reject the connection. Default: do not wait\n"
    "   for protocol violations.\n"
    "\n"
    "-g Use given directory as a base for graylist files. Each recipient must have a\n"
    "   subdirectory named with their domain (e.g. /.../graylist/foo.com) or\n"
    "   graylisting will not take place for that recipient. Default: do not graylist.\n"
    "\n"
    "-k Check the remote server's rDNS name for its IP address. If it is found AND\n"
    "   the rDNS name contains a keyword found in RDNS_KEYWORD_FILE, reject the\n"
    "   connection. Default: do not check the rDNS name for the IP address.\n"
    "\n"
    "-l Log activity to syslog. Very nice for figuring out why a sender is blocked.\n"
    "   Default: do not log activity, only log errors.\n"
    "\n"
    "-L Log all traffic to log files in LOG_DIR directory.\n"
    "\n"
    "-m Require a graylist entry to be present for MIN_SECS seconds before passing.\n"
    "   This option does nothing without -g. Default: allow a graylist entry to\n"
    "   pass with no wait period.\n"
    "\n"
    "-M Expire graylist entries after MAX_SECS seconds. This option does nothing\n"
    "   without -g. Default: never expire graylist entries.\n"
    "\n"
    "-r Reject the connection if no rDNS name exists for the remote server.\n"
    "   Default: do not check for an rDNS name.\n"
    "\n"
    "-R Reject the connection if the remote server's rDNS name does not resolve.\n"
    "   Default: do not attempt to resolve the rDNS name.\n"
    "\n"
    "-s Reject the sender address if it is found in the SENDER_BLACKLIST_FILE file.\n"
    "\n"
    "-t Forcibly disconnect after a total of MAX_CONNECTION_SECONDS seconds,\n"
    "   regardless of activity. A value of 0 disables this feature. Default: %d.\n"
    "\n"
    "-T Forcibly disconnect after MAX_COMMAND_SECONDS seconds of inactivity.\n"
    "   A value of 0 disables this feature. Default: %d.\n"
    "\n"
    "-w Search RDNS_WHITELIST_FILE for the base domain name of the remote server's\n"
    "   rDNS name. If it is found, accept the connection.\n"
    "\n"
    "-W Search IP_WHITELIST_FILE for the IP address of the remote server.  If it is\n"
    "   found, accept the connection.\n"
    "\n"
    "-u Append the given POLICY_LOCATION (typically a URL) to the rejection message\n"
    "   to explain why the rejection occurred. Don't rely on this -- many servers\n"
    "   (especially Exchange) don't put the rejection message in the bounce message\n"
    "   and most users don't read them anyway. Maximum %d characters.\n"
    "\n"
    "-x Check the remote server's IP address against realtime blackhole list RBL\n"
    "   and reject the connection if it is found.\n"
    "\n"
    "NOTE: Every option except a, g, m, M, t and T can be given multiple times to make\n"
    "multiple checks with different values.  If a, g, m, M, t or T are given multiple\n"
    "times, only the last value will be used.\n"
    , DEFAULT_TIMEOUT_CONNECTION_SECS
    , DEFAULT_TIMEOUT_COMMAND_SECS
    , MAX_POLICY
    );

  exit(0);

  return;
  }

char *find_environment_variable(char **environment, char *target_variable, int strlen_target_variable)
  {
  char *return_value;
  int i;

  return_value = NULL;

  if (environment != NULL)
    for (i = 0; environment[i] != NULL; i++)
      if (strncmp(environment[i], target_variable, strlen_target_variable) == 0)
        {
        return_value = environment[i] + strlen_target_variable + 1;
        break;
        }

  return(return_value);
  }

int find_address(char *haystack, int strlen_haystack, char **return_needle, int *strlen_username)
  {
  int return_value;
  int found_username;
  char *target_address;

  return_value = 0;
  found_username = 0;

  if ((haystack != NULL) &&
      (strlen_haystack > 0) &&
      (return_needle != NULL))
    {
    for (target_address = haystack; (target_address < (haystack + strlen_haystack)) && (target_address[0] != ':'); target_address++);
    for (; (target_address < (haystack + strlen_haystack)) && !isalnum(target_address[0]); target_address++);

    for (return_value = 0; ((target_address + return_value) < (haystack + strlen_haystack)); return_value++)
      if (!isalnum(target_address[return_value]))
        {
        switch (target_address[return_value])
          {
          case '@':
            if (strlen_username != NULL)
              *strlen_username = return_value;

            found_username = 1;

          case ':':
          case ',':
          case '.':
          case '#':
          case '[':
          case ']':
          case '-':
          case '"':
          case '\\':
          case '_':
            continue;
            break;
          }

        break;
        }

    *return_needle = (return_value > 0) ? target_address : NULL;
    if ((strlen_username != NULL) &&
        !found_username)
      *strlen_username = return_value;
    }

  return(return_value);
  }

char *canonicalize_path(char *destination_buf, int maxlen_destination_buf, char *target_path, int strlen_target_path)
  {
  char *filename_alphabet = ALPHABET_FILENAME;
  int i;

  i = 0;

  if ((destination_buf != NULL) &&
      (maxlen_destination_buf > 0))
    {
    if (target_path != NULL)
      for (i = 0; ((i < maxlen_destination_buf) && (i < strlen_target_path)); i++)
        destination_buf[i] = (strchr(filename_alphabet, target_path[i]) != NULL) ? target_path[i] : REPLACEMENT_FILENAME;

    destination_buf[i] = '\0';
    }

  return(destination_buf);
  }

int search_file(char *search_filename, char *target_string, int strlen_target_string, char wildcard)
  {
  int return_value;
  FILE *tmp_file;
  char tmp_buf[65536];
  int i;
  char *tmp_string;

  return_value = 0;

  if ((target_string != NULL) &&
      (strlen_target_string > 0))
    {
    if ((tmp_file = fopen(search_filename, "r")) != NULL)
      {
      i = 0;

      while (!feof(tmp_file) &&
             (i < 65536))
        {
        if (((fscanf(tmp_file, "%65535[^\r\n]%*[\r\n]", tmp_buf) == 1) ||
             (fscanf(tmp_file, "%*[\r\n]") == 0)) &&
            (tmp_buf[0] != '\0') &&
            (tmp_buf[0] != '#') &&
            ((tmp_string = strstr(target_string, tmp_buf)) != NULL) &&
            ((wildcard == '\0') ||
             (tmp_string == target_string) ||
             (tmp_buf[0] == '.')) &&
            (((tmp_string - target_string) + strlen(tmp_buf)) == strlen_target_string))
          {
          return_value = 1;
          break;
          }

        i++;
        }

      fclose(tmp_file);
      }
    else
      {
      syslog(LOG_ERR, SYSLOG_ERROR_OPEN_SEARCH "%s", search_filename);
      return_value = -1;
      }
    }

  return(return_value);
  }

int log_write(struct filter_settings *current_settings, int action, int target_fd, char *data, int data_length)
  {
  static FILE *log_file = NULL;
  static int last_action = -1;
  static char *prefixes[] = LOG_ACTION_PREFIX;
  char log_filename[MAX_BUF + 1];
  struct tm *tmp_tm;
  long tmp_time;

  if (action != LOG_ACTION_NONE)
    {
    tmp_tm = NULL;

    if ((current_settings == NULL) &&
        (log_file != NULL))
      {
      tmp_time = time(NULL);
      tmp_tm = localtime(&tmp_time);

      fprintf(log_file, "%.2d/%.2d/%d %.2d:%.2d:%.2d CLOSED\n", tmp_tm->tm_mon + 1, tmp_tm->tm_mday, tmp_tm->tm_year + 1900, tmp_tm->tm_hour, tmp_tm->tm_min, tmp_tm->tm_sec);

      fclose(log_file);
      log_file = NULL;
      }
    else if ((current_settings != NULL) &&
             (current_settings->log_dir != NULL) &&
             (log_file == NULL))
      {
      tmp_time = time(NULL);
      tmp_tm = localtime(&tmp_time);

      if ((current_settings->server_name != NULL) &&
          (current_settings->server_name[0] != '\0'))
        snprintf(log_filename, MAX_BUF, "%s/%d%.2d%.2d_%.2d%.2d%.2d_%s_%s", current_settings->log_dir, tmp_tm->tm_year + 1900, tmp_tm->tm_mon + 1, tmp_tm->tm_mday, tmp_tm->tm_hour, tmp_tm->tm_min, tmp_tm->tm_sec, current_settings->server_ip, current_settings->server_name);
      else
        snprintf(log_filename, MAX_BUF, "%s/%d%.2d%.2d_%.2d%.2d%.2d_%s", current_settings->log_dir, tmp_tm->tm_year + 1900, tmp_tm->tm_mon + 1, tmp_tm->tm_mday, tmp_tm->tm_hour, tmp_tm->tm_min, tmp_tm->tm_sec, current_settings->server_ip);

      if ((log_file = fopen(log_filename, "w")) != NULL)
        fprintf(log_file, "%.2d/%.2d/%d %.2d:%.2d:%.2d STARTED: To remote host = %s; to child process = %s\n", tmp_tm->tm_mon + 1, tmp_tm->tm_mday, tmp_tm->tm_year + 1900, tmp_tm->tm_hour, tmp_tm->tm_min, tmp_tm->tm_sec, prefixes[LOG_ACTION_CHILD_FROM], prefixes[LOG_ACTION_REMOTE_FROM]);
      else
        {
        syslog(LOG_ERR, SYSLOG_ERROR_OPEN_LOG "%s", log_filename);
        current_settings->log_dir = NULL;
        }
      }

    if (log_file != NULL)
      {
      if (tmp_tm == NULL)
        {
        tmp_time = time(NULL);
        tmp_tm = localtime(&tmp_time);
        }

      if (action != last_action)
        {
        fprintf(log_file, "%s %.2d/%.2d/%d %.2d:%.2d:%.2d\n", prefixes[action], tmp_tm->tm_mon + 1, tmp_tm->tm_mday, tmp_tm->tm_year + 1900, tmp_tm->tm_hour, tmp_tm->tm_min, tmp_tm->tm_sec);
        last_action = action;
        }

      fprintf(log_file, "%.*s", data_length, data);
      fflush(NULL);
      }
    }

  return((target_fd >= 0) ? write(target_fd, data, data_length) : 0);
  }

int smtp_filter(int outbound_fd, char *input_line, int strlen_input_line, struct filter_settings *current_settings)
  {
  int return_value;
  int i;
  char *target_address;
  int strlen_target_address;
  int strlen_username;
  struct stat tmp_stat;
  FILE *tmp_file;
  char log_entry[MAX_BUF + 1];
  int continue_processing;
  char tmp_recipient_domain[MAX_BUF + 1];
  char tmp_recipient_username[MAX_BUF + 1];
  char tmp_sender_address[MAX_BUF + 1];
  int search_results;

  return_value = FILTER_PASS;

  if (current_settings->inside_data)
    {
    if ((strlen_input_line >= STRLEN_SMTP_DATA_END) &&
        (strncasecmp(SMTP_DATA_END, input_line, STRLEN_SMTP_DATA_END) == 0) &&
        ((input_line[STRLEN_SMTP_DATA_END] == '\r') ||
         (input_line[STRLEN_SMTP_DATA_END] == '\n')))
      current_settings->inside_data = 0;
    }
  else if (((strlen_input_line >= STRLEN_SMTP_HELO_1) &&
            (strncasecmp(SMTP_HELO_1, input_line, STRLEN_SMTP_HELO_1) == 0)) ||
           ((strlen_input_line >= STRLEN_SMTP_HELO_2) &&
            (strncasecmp(SMTP_HELO_2, input_line, STRLEN_SMTP_HELO_2) == 0)))
    {
    if (current_settings->reject_message != NULL)
      {
      log_write(current_settings, LOG_ACTION_FILTER_FROM, outbound_fd, REJECT_SEVERITY_NONE, STRLEN_REJECT_SEVERITY);
      log_write(current_settings, LOG_ACTION_FILTER_FROM, outbound_fd, current_settings->reject_message, current_settings->strlen_reject_message);
      if (current_settings->policy_location != NULL)
        {
        log_write(current_settings, LOG_ACTION_FILTER_FROM, outbound_fd, ERROR_URL, STRLEN_ERROR_URL);
        log_write(current_settings, LOG_ACTION_FILTER_FROM, outbound_fd, current_settings->policy_location, current_settings->strlen_policy_location);
        }
      log_write(current_settings, LOG_ACTION_FILTER_FROM, outbound_fd, REJECT_CRLF, STRLEN_REJECT_CRLF);
      return_value = FILTER_INTERCEPT;
      }
    }
  else if ((strlen_input_line >= STRLEN_SMTP_MAIL_FROM) &&
           (strncasecmp(SMTP_MAIL_FROM, input_line, STRLEN_SMTP_MAIL_FROM) == 0))
    {
    if (((strlen_target_address = find_address(input_line, strlen_input_line, &target_address, &strlen_username)) > 0) &&
        (strlen_target_address < MAX_ADDRESS))
      {
      strncpy(current_settings->sender_address, target_address, strlen_target_address);
      current_settings->sender_address[strlen_target_address] = '\0';

      for (i = 0; i < strlen_target_address; i++)
        current_settings->sender_address[i] = tolower(current_settings->sender_address[i]);

      if ((current_settings->blacklist_sender_path == NULL) ||
          (search_file(current_settings->blacklist_sender_path, current_settings->sender_address, strlen_target_address, '\0') != 1))
        current_settings->local_sender = ((current_settings->local_domains_path == NULL) || (search_file(current_settings->local_domains_path, target_address + strlen_username + 1, (strlen_target_address - strlen_username) - 1, '.') == 1)) ? 1 : 0;
      else
        {
        current_settings->reject_message = ERROR_SENDER_BLACKLISTED;
        current_settings->strlen_reject_message = STRLEN_ERROR_SENDER_BLACKLISTED;
        current_settings->short_reject_message = SHORT_ERROR_SENDER_BLACKLISTED;
        }
      }

    if (current_settings->reject_message != NULL)
      {
      log_write(current_settings, LOG_ACTION_FILTER_FROM, outbound_fd, REJECT_SEVERITY_NONE, STRLEN_REJECT_SEVERITY);
      log_write(current_settings, LOG_ACTION_FILTER_FROM, outbound_fd, current_settings->reject_message, current_settings->strlen_reject_message);
      if (current_settings->policy_location != NULL)
        {
        log_write(current_settings, LOG_ACTION_FILTER_FROM, outbound_fd, ERROR_URL, STRLEN_ERROR_URL);
        log_write(current_settings, LOG_ACTION_FILTER_FROM, outbound_fd, current_settings->policy_location, current_settings->strlen_policy_location);
        }
      log_write(current_settings, LOG_ACTION_FILTER_FROM, outbound_fd, REJECT_CRLF, STRLEN_REJECT_CRLF);
      return_value = FILTER_INTERCEPT;
      }
    }
  else if ((strlen_input_line >= STRLEN_SMTP_RCPT_TO) &&
           (strncasecmp(SMTP_RCPT_TO, input_line, STRLEN_SMTP_RCPT_TO) == 0))
    {
    if ((strlen_target_address = find_address(input_line, strlen_input_line, &target_address, &strlen_username)) > 0)
      {
      strncpy(current_settings->recipient_address, target_address, strlen_target_address);
      current_settings->recipient_address[strlen_target_address] = '\0';

      for (i = 0; i < strlen_target_address; i++)
        current_settings->recipient_address[i] = tolower(current_settings->recipient_address[i]);
      }

    if (current_settings->reject_message != NULL)
      {
      log_write(current_settings, LOG_ACTION_FILTER_FROM, outbound_fd, REJECT_SEVERITY_TEMPORARY, STRLEN_REJECT_SEVERITY);
      log_write(current_settings, LOG_ACTION_FILTER_FROM, outbound_fd, current_settings->reject_message, current_settings->strlen_reject_message);
      if (current_settings->policy_location != NULL)
        {
        log_write(current_settings, LOG_ACTION_FILTER_FROM, outbound_fd, ERROR_URL, STRLEN_ERROR_URL);
        log_write(current_settings, LOG_ACTION_FILTER_FROM, outbound_fd, current_settings->policy_location, current_settings->strlen_policy_location);
        }
      log_write(current_settings, LOG_ACTION_FILTER_FROM, outbound_fd, REJECT_CRLF, STRLEN_REJECT_CRLF);
      return_value = FILTER_INTERCEPT;
      }
    else if (strlen_username == strlen_target_address)
      {
      log_write(current_settings, LOG_ACTION_FILTER_FROM, outbound_fd, REJECT_SEVERITY_TEMPORARY, STRLEN_REJECT_SEVERITY);
      log_write(current_settings, LOG_ACTION_FILTER_FROM, outbound_fd, ERROR_RCPT_TO_LOCAL, STRLEN_ERROR_RCPT_TO_LOCAL);
      if (current_settings->policy_location != NULL)
        {
        log_write(current_settings, LOG_ACTION_FILTER_FROM, outbound_fd, ERROR_URL, STRLEN_ERROR_URL);
        log_write(current_settings, LOG_ACTION_FILTER_FROM, outbound_fd, current_settings->policy_location, current_settings->strlen_policy_location);
        }
      log_write(current_settings, LOG_ACTION_FILTER_FROM, outbound_fd, REJECT_CRLF, STRLEN_REJECT_CRLF);
      current_settings->transient_reject_message = SHORT_ERROR_RCPT_TO_LOCAL;
      return_value = FILTER_INTERCEPT;
      }
    else if ((current_settings->max_rcpt_to > 0) &&
        !current_settings->local_sender &&
        (current_settings->num_rcpt_to >= current_settings->max_rcpt_to))
      {
      log_write(current_settings, LOG_ACTION_FILTER_FROM, outbound_fd, REJECT_SEVERITY_TEMPORARY, STRLEN_REJECT_SEVERITY);
      log_write(current_settings, LOG_ACTION_FILTER_FROM, outbound_fd, ERROR_RCPT_TO, STRLEN_ERROR_RCPT_TO);
      if (current_settings->policy_location != NULL)
        {
        log_write(current_settings, LOG_ACTION_FILTER_FROM, outbound_fd, ERROR_URL, STRLEN_ERROR_URL);
        log_write(current_settings, LOG_ACTION_FILTER_FROM, outbound_fd, current_settings->policy_location, current_settings->strlen_policy_location);
        }
      log_write(current_settings, LOG_ACTION_FILTER_FROM, outbound_fd, REJECT_CRLF, STRLEN_REJECT_CRLF);
      current_settings->transient_reject_message = SHORT_ERROR_RCPT_TO;
      return_value = FILTER_INTERCEPT;
      }
    // Construct the path to the recipient's domain's graylist folder and check if it exists
    else if ((current_settings->graylist_dir != NULL) &&
             (strlen_target_address > 0) &&
             (snprintf(current_settings->graylist_path, MAX_PATH, "%s/%s", current_settings->graylist_dir, canonicalize_path(tmp_recipient_domain, MAX_PATH, target_address + strlen_username + 1, (strlen_target_address - strlen_username) - 1)) < MAX_PATH) &&
             (stat(current_settings->graylist_path, &tmp_stat) == 0) &&
             (S_ISDIR(tmp_stat.st_mode)))
      {
      continue_processing = 1;
      if ((snprintf(current_settings->graylist_path, MAX_PATH, "%s/%s/%s", current_settings->graylist_dir, tmp_recipient_domain, canonicalize_path(tmp_recipient_username, MAX_PATH, target_address, strlen_username)) < MAX_PATH) &&
          (stat(current_settings->graylist_path, &tmp_stat) != 0))
        {
        if (errno == ENOENT)
          {
          if (mkdir(current_settings->graylist_path, MKDIR_MODE) != 0)
            {
            continue_processing = 0;
            syslog(LOG_ERR, SYSLOG_ERROR_MKDIR "%s: %s", current_settings->graylist_path, strerror(errno));
            }
          }
        else
          {
          continue_processing = 0;
          syslog(LOG_ERR, SYSLOG_ERROR_STAT "%s: %s", current_settings->graylist_path, strerror(errno));
          }
        }

      // Construct the path to the sender->user file and check if it exists
      if (continue_processing &&
          (snprintf(current_settings->graylist_path, MAX_PATH, "%s/%s/%s/%s", current_settings->graylist_dir, tmp_recipient_domain, tmp_recipient_username, (current_settings->sender_address[0] != '\0') ? canonicalize_path(tmp_sender_address, MAX_PATH, current_settings->sender_address, strlen(current_settings->sender_address)) : SENDER_ADDRESS_NONE) < MAX_PATH) &&
          (stat(current_settings->graylist_path, &tmp_stat) == 0))
        // Check that the file is a "regular" file and the age matches the command line options
        if ((S_ISREG(tmp_stat.st_mode)) &&
            (((current_settings->graylist_min_secs > 0) &&
              (((tmp_stat.st_mtime + current_settings->graylist_min_secs) <= time(NULL)) ||
               (tmp_stat.st_size > 0))) ||
             (current_settings->graylist_min_secs == 0)) &&
            (((current_settings->graylist_max_secs > 0) &&
              ((tmp_stat.st_mtime + current_settings->graylist_max_secs) >= time(NULL))) ||
             (current_settings->graylist_max_secs == 0)))
          {
          // User was graylisted but is now allowed.  Log the name of their sending host to the file.
          if ((current_settings->server_name != NULL) &&
              (current_settings->server_name[0] != '\0'))
            snprintf(log_entry, MAX_BUF, "%s %s", current_settings->server_ip, current_settings->server_name);
          else
            snprintf(log_entry, MAX_BUF, "%s", current_settings->server_ip);

          search_results = search_file(current_settings->graylist_path, log_entry, strlen(log_entry), '\0');
          if ((tmp_file = fopen(current_settings->graylist_path, "a")) != NULL)
            {
            if (search_results != 1)
              fprintf(tmp_file, "%s\n", log_entry);

            fclose(tmp_file);
            }
          else
            syslog(LOG_ERR, SYSLOG_ERROR_GRAYLIST_FILE "%s", current_settings->graylist_path);
          }
        else
          {
          // The file's age is outside the bounds of the command line options.
          return_value = FILTER_INTERCEPT;

          // The file was too old -- truncate it.
          if ((current_settings->graylist_max_secs > 0) &&
              ((tmp_stat.st_mtime + current_settings->graylist_max_secs) <= time(NULL)))
            {
            if ((tmp_file = fopen(current_settings->graylist_path, "w")) != NULL)
              fclose(tmp_file);
            else
              {
              return_value = FILTER_PASS;
              syslog(LOG_ERR, SYSLOG_ERROR_GRAYLIST_FILE "%s", current_settings->graylist_path);
              }
            }

          if (return_value != FILTER_PASS)
            {
            log_write(current_settings, LOG_ACTION_FILTER_FROM, outbound_fd, REJECT_SEVERITY_TEMPORARY, STRLEN_REJECT_SEVERITY);
            log_write(current_settings, LOG_ACTION_FILTER_FROM, outbound_fd, ERROR_GRAYLISTED, STRLEN_ERROR_GRAYLISTED);
            if (current_settings->policy_location != NULL)
              {
              log_write(current_settings, LOG_ACTION_FILTER_FROM, outbound_fd, ERROR_URL, STRLEN_ERROR_URL);
              log_write(current_settings, LOG_ACTION_FILTER_FROM, outbound_fd, current_settings->policy_location, current_settings->strlen_policy_location);
              }
            log_write(current_settings, LOG_ACTION_FILTER_FROM, outbound_fd, REJECT_CRLF, STRLEN_REJECT_CRLF);
            current_settings->transient_reject_message = SHORT_ERROR_GRAYLISTED;
            }
          }
      // Sender->user graylist file does not exist -- create it.
      else if (continue_processing &&
               ((tmp_file = fopen(current_settings->graylist_path, "w")) != NULL))
        {
        fclose(tmp_file);

        log_write(current_settings, LOG_ACTION_FILTER_FROM, outbound_fd, REJECT_SEVERITY_TEMPORARY, STRLEN_REJECT_SEVERITY);
        log_write(current_settings, LOG_ACTION_FILTER_FROM, outbound_fd, ERROR_GRAYLISTED, STRLEN_ERROR_GRAYLISTED);
        if (current_settings->policy_location != NULL)
          {
          log_write(current_settings, LOG_ACTION_FILTER_FROM, outbound_fd, ERROR_URL, STRLEN_ERROR_URL);
          log_write(current_settings, LOG_ACTION_FILTER_FROM, outbound_fd, current_settings->policy_location, current_settings->strlen_policy_location);
          }
        log_write(current_settings, LOG_ACTION_FILTER_FROM, outbound_fd, REJECT_CRLF, STRLEN_REJECT_CRLF);
        current_settings->transient_reject_message = SHORT_ERROR_GRAYLISTED;

        return_value = FILTER_INTERCEPT;
        }
      else if (continue_processing)
        syslog(LOG_ERR, SYSLOG_ERROR_GRAYLIST_FILE "%s", current_settings->graylist_path);
      }

    if (current_settings->use_syslog)
      syslog(LOG_INFO, "%s from: %s to: %s origin: %s %s", (current_settings->short_reject_message != NULL) ? current_settings->short_reject_message : ((current_settings->transient_reject_message != NULL) ? current_settings->transient_reject_message : SHORT_SUCCESS), (current_settings->sender_address[0] != '\0') ? current_settings->sender_address : SYSLOG_MISSING_DATA, (current_settings->recipient_address[0] != '\0') ? current_settings->recipient_address : SYSLOG_MISSING_DATA, current_settings->server_ip, ((current_settings->server_name != NULL) && (current_settings->server_name[0] != '\0')) ? current_settings->server_name : SYSLOG_MISSING_DATA);

    current_settings->transient_reject_message = NULL;

    if (return_value == FILTER_PASS)
      current_settings->num_rcpt_to++;
    }
  else if ((strlen_input_line >= STRLEN_SMTP_DATA) &&
           (strncasecmp(SMTP_DATA, input_line, STRLEN_SMTP_DATA) == 0))
    {
    if (current_settings->reject_message == NULL)
      current_settings->inside_data = 1;
    else
      {
      log_write(current_settings, LOG_ACTION_FILTER_FROM, outbound_fd, REJECT_SEVERITY_TEMPORARY, STRLEN_REJECT_SEVERITY);
      log_write(current_settings, LOG_ACTION_FILTER_FROM, outbound_fd, current_settings->reject_message, current_settings->strlen_reject_message);
      if (current_settings->policy_location != NULL)
        {
        log_write(current_settings, LOG_ACTION_FILTER_FROM, outbound_fd, ERROR_URL, STRLEN_ERROR_URL);
        log_write(current_settings, LOG_ACTION_FILTER_FROM, outbound_fd, current_settings->policy_location, current_settings->strlen_policy_location);
        }
      log_write(current_settings, LOG_ACTION_FILTER_FROM, outbound_fd, REJECT_CRLF, STRLEN_REJECT_CRLF);
      return_value = FILTER_INTERCEPT;
      }
    }
  else if ((strlen_input_line >= STRLEN_SMTP_QUIT) &&
           (strncasecmp(SMTP_QUIT, input_line, STRLEN_SMTP_QUIT) == 0))
    {
    if (current_settings->reject_message != NULL)
      {
      log_write(current_settings, LOG_ACTION_FILTER_FROM, outbound_fd, REJECT_SEVERITY_QUIT, STRLEN_REJECT_SEVERITY);
      log_write(current_settings, LOG_ACTION_FILTER_FROM, outbound_fd, current_settings->reject_message, current_settings->strlen_reject_message);
      if (current_settings->policy_location != NULL)
        {
        log_write(current_settings, LOG_ACTION_FILTER_FROM, outbound_fd, ERROR_URL, STRLEN_ERROR_URL);
        log_write(current_settings, LOG_ACTION_FILTER_FROM, outbound_fd, current_settings->policy_location, current_settings->strlen_policy_location);
        }
      log_write(current_settings, LOG_ACTION_FILTER_FROM, outbound_fd, REJECT_CRLF, STRLEN_REJECT_CRLF);
      return_value = FILTER_QUIT;
      }
    }
  else if ((strlen_input_line >= STRLEN_SMTP_RSET) &&
           (strncasecmp(SMTP_RSET, input_line, STRLEN_SMTP_RSET) == 0))
    {
    if (current_settings->reject_message == NULL)
      {
      current_settings->num_rcpt_to = 0;
      current_settings->local_sender = 1;
      current_settings->sender_address[0] = '\0';
      current_settings->recipient_address[0] = '\0';
      current_settings->graylist_path[0] = '\0';
      }
    else
      {
      log_write(current_settings, LOG_ACTION_FILTER_FROM, outbound_fd, REJECT_SEVERITY_PERMANENT, STRLEN_REJECT_SEVERITY);
      log_write(current_settings, LOG_ACTION_FILTER_FROM, outbound_fd, current_settings->reject_message, current_settings->strlen_reject_message);
      if (current_settings->policy_location != NULL)
        {
        log_write(current_settings, LOG_ACTION_FILTER_FROM, outbound_fd, ERROR_URL, STRLEN_ERROR_URL);
        log_write(current_settings, LOG_ACTION_FILTER_FROM, outbound_fd, current_settings->policy_location, current_settings->strlen_policy_location);
        }
      log_write(current_settings, LOG_ACTION_FILTER_FROM, outbound_fd, REJECT_CRLF, STRLEN_REJECT_CRLF);
      return_value = FILTER_INTERCEPT;
      }
    }
  else if ((strlen_input_line < STRLEN_SMTP_QUIT) ||
           (strncasecmp(SMTP_QUIT, input_line, STRLEN_SMTP_QUIT) != 0))
    {
    if (current_settings->reject_message != NULL)
      {
      log_write(current_settings, LOG_ACTION_FILTER_FROM, outbound_fd, REJECT_SEVERITY_PERMANENT, STRLEN_REJECT_SEVERITY);
      log_write(current_settings, LOG_ACTION_FILTER_FROM, outbound_fd, current_settings->reject_message, current_settings->strlen_reject_message);
      if (current_settings->policy_location != NULL)
        {
        log_write(current_settings, LOG_ACTION_FILTER_FROM, outbound_fd, ERROR_URL, STRLEN_ERROR_URL);
        log_write(current_settings, LOG_ACTION_FILTER_FROM, outbound_fd, current_settings->policy_location, current_settings->strlen_policy_location);
        }
      log_write(current_settings, LOG_ACTION_FILTER_FROM, outbound_fd, REJECT_CRLF, STRLEN_REJECT_CRLF);
      return_value = FILTER_INTERCEPT;
      }
    }

  return(return_value);
  }

void middleman(int *inbound_fd, int *outbound_fd, char line_terminator, struct filter_settings *current_settings)
  {
  char tmp_char;
  char buf_input[MAX_BUF + 1];
  int strlen_buf_input;
  fd_set listen_set;
  struct timeval listen_timeout;
  int more_stdinput;
  int more_childinput;
  int discard_childinput;
  long connection_start;
  long command_start;
  long current_time;
  int filter_return;

  strlen_buf_input = 0;

  more_stdinput = 1;
  more_childinput = ((*inbound_fd >= 0) && (*outbound_fd >= 0)) ? 1 : 0;
  discard_childinput = !more_childinput;

  connection_start = time(NULL);
  command_start = time(NULL);

  while ((more_childinput ||
         discard_childinput) &&
         more_stdinput)
    {
    current_time = time(NULL);

    if (((current_settings->timeout_connection > 0) &&
         ((current_time - connection_start) > current_settings->timeout_connection)) ||
        ((current_settings->timeout_command > 0) &&
         ((current_time - command_start) > current_settings->timeout_command)))
      {
      if (*outbound_fd >= 0)
        {
        log_write(current_settings, LOG_ACTION_FILTER_TO, *outbound_fd, CHILD_QUIT, STRLEN_CHILD_QUIT);
        close(*outbound_fd);
        *outbound_fd = -1;
        }

      log_write(current_settings, LOG_ACTION_FILTER_FROM, STDOUT_FD, REJECT_SEVERITY_TEMPORARY, STRLEN_REJECT_SEVERITY);
      log_write(current_settings, LOG_ACTION_FILTER_FROM, STDOUT_FD, ERROR_TIMEOUT, STRLEN_ERROR_TIMEOUT);
      if (current_settings->policy_location != NULL)
        {
        log_write(current_settings, LOG_ACTION_FILTER_FROM, STDOUT_FD, ERROR_URL, STRLEN_ERROR_URL);
        log_write(current_settings, LOG_ACTION_FILTER_FROM, STDOUT_FD, current_settings->policy_location, current_settings->strlen_policy_location);
        }
      log_write(current_settings, LOG_ACTION_FILTER_FROM, STDOUT_FD, REJECT_CRLF, STRLEN_REJECT_CRLF);

      more_stdinput = 0;
      discard_childinput = 1;

      if (current_settings->use_syslog)
        syslog(LOG_INFO, "%s from: %s to: %s origin: %s %s %s", SHORT_ERROR_TIMEOUT, (current_settings->sender_address[0] != '\0') ? current_settings->sender_address : SYSLOG_MISSING_DATA, (current_settings->recipient_address[0] != '\0') ? current_settings->recipient_address : SYSLOG_MISSING_DATA, current_settings->server_ip, ((current_settings->server_name != NULL) && (current_settings->server_name[0] != '\0')) ? current_settings->server_name : SYSLOG_MISSING_DATA, (current_settings->short_reject_message != NULL) ? current_settings->short_reject_message : SYSLOG_MISSING_DATA);
      }

    listen_timeout.tv_sec = 0;
    listen_timeout.tv_usec = 0;

    if ((current_settings->timeout_connection > 0) ||
        (current_settings->timeout_command > 0))
      {
      if (current_settings->timeout_command > 0)
        listen_timeout.tv_sec = MAXVAL(0, current_settings->timeout_command - (current_time - command_start));

      if (current_settings->timeout_connection > 0)
        listen_timeout.tv_sec = MAXVAL(0, MINVAL(listen_timeout.tv_sec, current_settings->timeout_connection - (current_time - connection_start)));

      if (listen_timeout.tv_sec == 0)
        listen_timeout.tv_usec = 500000;
      }
    else
      listen_timeout.tv_sec = 10;

    FD_ZERO(&listen_set);
    if (more_stdinput)
      FD_SET(STDIN_FD, &listen_set);
    if (more_childinput)
      FD_SET(*inbound_fd, &listen_set);

    if (select(MAXVAL(STDIN_FD, *inbound_fd) + 1, &listen_set, NULL, NULL, &listen_timeout) > 0)
      {
      if ((*inbound_fd >= 0) &&
          FD_ISSET(*inbound_fd, &listen_set))
        {
        if (read(*inbound_fd, &tmp_char, 1) > 0)
          {
          if (!discard_childinput)
            log_write(current_settings, LOG_ACTION_CHILD_FROM, STDOUT_FD, &tmp_char, 1);
          else
            log_write(current_settings, LOG_ACTION_CHILD_FROM_DISCARDED, -1, &tmp_char, 1);
          }
        else
          {
          more_childinput = 0;
          close(*inbound_fd);
          *inbound_fd = -1;
          }
        }

      if (FD_ISSET(STDIN_FD, &listen_set))
        {
        if (read(STDIN_FD, &tmp_char, 1) > 0)
          {
          command_start = time(NULL);

          if (line_terminator != '\0')
            {
            buf_input[strlen_buf_input] = tmp_char;
            strlen_buf_input++;

            if ((tmp_char == line_terminator) ||
                (strlen_buf_input >= MAX_BUF))
              {
              log_write(current_settings, LOG_ACTION_REMOTE_FROM, -1, buf_input, strlen_buf_input);

              filter_return = (tmp_char == line_terminator) ? smtp_filter(STDOUT_FD, buf_input, strlen_buf_input, current_settings) : FILTER_PASS;
              switch (filter_return)
                {
                case FILTER_PASS:
                  log_write(current_settings, LOG_ACTION_NONE, *outbound_fd, buf_input, strlen_buf_input);
                  break;
                case FILTER_QUIT:
                  more_stdinput = 0;
                  discard_childinput = 1;

                  break;
                case FILTER_INTERCEPT:
                  if ((current_settings->reject_message != NULL) &&
                      (*outbound_fd >= 0) &&
                      !discard_childinput)
                    {
                    log_write(current_settings, LOG_ACTION_FILTER_TO, *outbound_fd, CHILD_QUIT, STRLEN_CHILD_QUIT);
                    close(*outbound_fd);
                    *outbound_fd = -1;

                    discard_childinput = 1;
                    }
                }

              strlen_buf_input = 0;
              }
            }
          else
            log_write(current_settings, LOG_ACTION_REMOTE_FROM, *outbound_fd, &tmp_char, 1);
          }
        else
          {
          if (*outbound_fd >= 0)
            {
            if ((line_terminator != '\0') &&
                (strlen_buf_input > 0))
              {
              log_write(current_settings, LOG_ACTION_REMOTE_FROM, *outbound_fd, buf_input, strlen_buf_input);
              strlen_buf_input = 0;
              }

            close(*outbound_fd);
            *outbound_fd = -1;
            }

          more_stdinput = 0;
          }
        }
      }

    waitpid(-1, NULL, WNOHANG);
    }

  return;
  }

int search_ip(char *target_ip, int strlen_target_ip, char *target_name, int strlen_target_name)
  {
  int j;
  int k;

  j = 0;
  k = 0;
  while (k < strlen_target_name)
    if (j < strlen_target_ip)
      if ((target_ip[j] >= '0') &&
          (target_ip[j] <= '9'))
        {
        if (target_ip[j] == target_name[k])
          j++;
        else
          j = 0;

        k++;
        }
      else
        {
        while ((k < strlen_target_name) &&
               ((target_name[k] < '0') ||
                (target_name[k] > '9')))
          k++;

        j++;
        }
    else
      break;

  return(((j > 0) && (j == strlen_target_ip)) ? 1 : 0);
  }

int is_ip_in_name(char *target_ip, char *target_name)
  {
  int return_value;
  char ip_octets[4][4];
  char tmp_ip[16];
  int tmp_strlen;
  struct in_addr tmp_addr;
  int strlen_target_ip;
  int strlen_target_name;

  return_value = -1;

  if ((target_ip != NULL) &&
      ((strlen_target_ip = strlen(target_ip)) > 0) &&
      (sscanf(target_ip, "%3[0-9].%3[0-9].%3[0-9].%3[0-9]", ip_octets[0], ip_octets[1], ip_octets[2], ip_octets[3]) == 4) &&
      (target_name != NULL) &&
      ((strlen_target_name = strlen(target_name)) > 0))
    {
    return_value = 0;

//printf("Looking for %s\n", target_ip); //FIXME
    // This block looks for the dotted-quad IP address in the rDNS name.
    return_value = search_ip(target_ip, strlen_target_ip, target_name, strlen_target_name);

    // This block looks for the zero-padded dotted-quad IP address in the rDNS
    // name.  For example, if the IP is 11.22.33.44, this block looks for
    // 011.022.033.044
    if (!return_value)
      {
      tmp_strlen = sprintf(tmp_ip, "%.3d.%.3d.%.3d.%.3d", atoi(ip_octets[0]), atoi(ip_octets[1]), atoi(ip_octets[2]), atoi(ip_octets[3]));
//printf("Looking for %s\n", tmp_ip); //FIXME
      return_value = search_ip(tmp_ip, tmp_strlen, target_name, strlen_target_name);
      }

    // This block looks for the reversed dotted-quad IP address in the rDNS
    // name.  For example, if the IP is 11.22.33.44, this block looks for
    // 44.33.22.11.
    if (!return_value)
      {
      tmp_strlen = sprintf(tmp_ip, "%s.%s.%s.%s", ip_octets[3], ip_octets[2], ip_octets[1], ip_octets[0]);
//printf("Looking for %s\n", tmp_ip); //FIXME
      return_value = search_ip(tmp_ip, tmp_strlen, target_name, strlen_target_name);
      }

    // This block looks for the slightly out-of-order dotted-quad IP address in the rDNS
    // name.  For example, if the IP is 11.22.33.44, this block looks for
    // 44.11.22.33
    if (!return_value)
      {
      tmp_strlen = sprintf(tmp_ip, "%s.%s.%s.%s", ip_octets[3], ip_octets[0], ip_octets[1], ip_octets[2]);
//printf("Looking for %s\n", tmp_ip); //FIXME
      return_value = search_ip(tmp_ip, tmp_strlen, target_name, strlen_target_name);
      }

    // This block looks for the strangely out-of-order dotted-quad IP address in the rDNS
    // name.  For example, if the IP is 11.22.33.44, this block looks for
    // 33.22.11.44
    if (!return_value)
      {
      tmp_strlen = sprintf(tmp_ip, "%s.%s.%s.%s", ip_octets[2], ip_octets[1], ip_octets[0], ip_octets[3]);
//printf("Looking for %s\n", tmp_ip); //FIXME
      return_value = search_ip(tmp_ip, tmp_strlen, target_name, strlen_target_name);
      }

    // This block looks for the strangely out-of-order dotted-quad IP address in the rDNS
    // name.  For example, if the IP is 11.22.33.44, this block looks for
    // 44.33.1122
    if (!return_value)
      {
      tmp_strlen = sprintf(tmp_ip, "%s.%s.%s%s", ip_octets[3], ip_octets[2], ip_octets[0], ip_octets[1]);
//printf("Looking for %s\n", tmp_ip); //FIXME
      return_value = search_ip(tmp_ip, tmp_strlen, target_name, strlen_target_name);
      }

    // This block looks for the strangely out-of-order dotted-quad IP address in the rDNS
    // name.  For example, if the IP is 11.22.33.44, this block looks for
    // 3344.11.22
    if (!return_value)
      {
      tmp_strlen = sprintf(tmp_ip, "%s%s.%s.%s", ip_octets[2], ip_octets[3], ip_octets[0], ip_octets[1]);
//printf("Looking for %s\n", tmp_ip); //FIXME
      return_value = search_ip(tmp_ip, tmp_strlen, target_name, strlen_target_name);
      }

    // This block looks for the dotted-quad IP address with the last two octets encoded as a long integer in the rDNS
    // name.  For example, if the IP is 11.22.33.44, this block looks for
    // 11.22.8492
    if (!return_value)
      {
      tmp_strlen = sprintf(tmp_ip, "%s.%s.%d", ip_octets[0], ip_octets[1], (atoi(ip_octets[2]) * 256) + atoi(ip_octets[3]));
//printf("Looking for %s\n", tmp_ip); //FIXME
      return_value = search_ip(tmp_ip, tmp_strlen, target_name, strlen_target_name);
      }

    // This block looks for the undotted-quad IP address in the rDNS name.
    // For example, if the IP is 11.22.33.44, this block looks for
    // 11223344.
    if (!return_value)
      {
      sprintf(tmp_ip, "%s%s%s%s", ip_octets[0], ip_octets[1], ip_octets[2], ip_octets[3]);
//printf("Looking for %s\n", tmp_ip); //FIXME

      if (strstr(target_name, tmp_ip) != NULL)
        return_value = 1;
      }

    // This block looks for the undotted-, zero-padded-quad IP address in the
    // rDNS name.  For example, if the IP is 11.22.33.44, this block looks for
    // 011022033044
    if (!return_value)
      {
      sprintf(tmp_ip, "%.3d%.3d%.3d%.3d", atoi(ip_octets[0]), atoi(ip_octets[1]), atoi(ip_octets[2]), atoi(ip_octets[3]));
//printf("Looking for %s\n", tmp_ip); //FIXME

      if (strstr(target_name, tmp_ip) != NULL)
        return_value = 1;
      }

    // This block looks for the undotted-, partially zero-padded-quad IP address in the
    // rDNS name.  For example, if the IP is 11.22.33.44, this block looks for
    // 11022033044
    if (!return_value)
      {
      sprintf(tmp_ip, "%s%.3d%.3d%.3d", ip_octets[0], atoi(ip_octets[1]), atoi(ip_octets[2]), atoi(ip_octets[3]));
//printf("Looking for %s\n", tmp_ip); //FIXME

      if (strstr(target_name, tmp_ip) != NULL)
        return_value = 1;
      }

    // This block looks for the undotted-, partially zero-padded-quad IP address in the
    // rDNS name.  For example, if the IP is 11.22.33.44, this block looks for
    // 1122033044
    if (!return_value)
      {
      sprintf(tmp_ip, "%s%s%.3d%.3d", ip_octets[0], ip_octets[1], atoi(ip_octets[2]), atoi(ip_octets[3]));
//printf("Looking for %s\n", tmp_ip); //FIXME

      if (strstr(target_name, tmp_ip) != NULL)
        return_value = 1;
      }

    // This block looks for the undotted-, partially zero-padded-quad IP address in the
    // rDNS name.  For example, if the IP is 11.22.33.44, this block looks for
    // 112233044
    if (!return_value)
      {
      sprintf(tmp_ip, "%s%s%s%.3d", ip_octets[0], ip_octets[1], ip_octets[2], atoi(ip_octets[3]));
//printf("Looking for %s\n", tmp_ip); //FIXME

      if (strstr(target_name, tmp_ip) != NULL)
        return_value = 1;
      }

    // This block looks for the undotted-quad reversed IP address in the rDNS name.
    // For example, if the IP is 11.22.33.44, this block looks for
    // 44332211.
    if (!return_value)
      {
      sprintf(tmp_ip, "%s%s%s%s", ip_octets[3], ip_octets[2], ip_octets[1], ip_octets[0]);
//printf("Looking for %s\n", tmp_ip); //FIXME

      if (strstr(target_name, tmp_ip) != NULL)
        return_value = 1;
      }

    // This block looks for the undotted-, zero-padded-quad reversed IP address in the
    // rDNS name.  For example, if the IP is 11.22.33.44, this block looks for
    // 044033022011
    if (!return_value)
      {
      sprintf(tmp_ip, "%.3d%.3d%.3d%.3d", atoi(ip_octets[3]), atoi(ip_octets[2]), atoi(ip_octets[1]), atoi(ip_octets[0]));
//printf("Looking for %s\n", tmp_ip); //FIXME

      if (strstr(target_name, tmp_ip) != NULL)
        return_value = 1;
      }

    // This block looks for the IP address as an unsigned long integer.
    // For example, if the IP is 85.135.72.234, this block looks for
    // 3930621781.
    if (!return_value &&
        inet_aton(target_ip, &tmp_addr))
      {
      sprintf(tmp_ip, "%lu", (long unsigned int)tmp_addr.s_addr);
//printf("Looking for %s\n", tmp_ip); //FIXME

      if (strstr(target_name, tmp_ip) != NULL)
        return_value = 1;
      }

    // This block looks for the IP address as a sequence of hex bytes.
    // For example, if the IP is 85.135.72.234, this block looks for
    // 5080d7e3.
    if (!return_value)
      {
      sprintf(tmp_ip, "%.2x%.2x%.2x%.2x", atoi(ip_octets[0]), atoi(ip_octets[1]), atoi(ip_octets[2]), atoi(ip_octets[3]));
//printf("Looking for %s\n", tmp_ip); //FIXME

      if (strstr(target_name, tmp_ip) != NULL)
        return_value = 1;
      }
    }

  return(return_value);
  }

int check_rdns_resolve(char *target_hostname)
  {
  int return_value;

  return_value = 0;

  if ((target_hostname != NULL) &&
      (target_hostname[0] != '\0') &&
      ((gethostbyname(target_hostname) != NULL) ||
       ((h_errno != HOST_NOT_FOUND) &&
        (h_errno != NO_ADDRESS) &&
        (h_errno != NO_DATA))))
    return_value = 1;

  return(return_value);
  }

int check_country_code(char *target_name, int strlen_target_name)
  {
  int return_value;
  int j;

  return_value = 0;

  if ((target_name != NULL) &&
      (strlen_target_name > 0))
    for (j = (strlen_target_name - 1); j >= 0; j--)
      if (target_name[j] == '.')
        {
        if ((strlen_target_name - j) == 3)
          return_value = 1;

        break;
        }

  return(return_value);
  }

int check_rdns_keywords(char *keywords_filename, char *target_name, int strlen_target_name)
  {
  int return_value;
  FILE *tmp_file;
  char tmp_buf[65536];
  int i;
  int j;
  char *tld_start;
  int strlen_keyword;
  int strlen_subkeyword;
  char *tmp_char;
  int found;
  int not_found;

  return_value = 0;

  if ((target_name != NULL) &&
      (strlen_target_name > 0) &&
      ((tmp_file = fopen(keywords_filename, "r")) != NULL))
    {
    // It's important to know where the domain name begins, since
    // disqualifying mail.dhcp.com isn't fair.
    tld_start = NULL;
    for (j = (strlen_target_name - 1); j >= 0; j--)
      if (target_name[j] == '.')
        {
        for (j--; j >= 0; j--)
          if (target_name[j] == '.')
            {
            tld_start = target_name + j + 1;
            break;
            }

        break;
        }

    if (tld_start != NULL)
      {
      i = 0;

      while (!feof(tmp_file) &&
             (i < 65536))
        {
        if ((fscanf(tmp_file, "%65535[^\r\n]%*[\r\n]", tmp_buf) == 1) &&
            (tmp_buf[0] != '\0') &&
            (tmp_buf[0] != '#') &&
            ((strlen_keyword = strlen(tmp_buf)) > 0))
          {
          for (j = 0; j < strlen_keyword; j++)
            if (tmp_buf[j] == ' ')
              tmp_buf[j] = '\0';

          j = 0;
          not_found = 0;
          while (j < strlen_keyword)
            {
            strlen_subkeyword = strlen(tmp_buf + j);

            found = 0;
            tmp_char = target_name;
            while (((tmp_char - target_name) < strlen_target_name) &&
                   ((tmp_char = strstr(tmp_char, tmp_buf + j)) != NULL) &&
                   (tmp_char < tld_start))
              {
              // The preceding and following characters around the match must be
              // non-alphanumeric so we don't reject 10.0.0.1.notdialup.example.com.
              // This means 10.0.0.1.dialup3.example.com will get through but the
              // alternative is worse.
              if (((tmp_char == target_name) ||
                   !isalnum((tmp_buf + j)[0]) ||
                   !isalnum((tmp_char - 1)[0])) &&
                  (!isalnum((tmp_buf + j + strlen_subkeyword - 1)[0]) ||
                   !isalnum((tmp_char + strlen_subkeyword)[0])))
                {
                found = 1;
                break;
                }

              tmp_char += strlen_subkeyword;
              }

            if (!found)
              {
              not_found = 1;
              break;
              }

            while (j < strlen_keyword)
              if (tmp_buf[j++] == '\0')
                break;
            }

          if (!not_found)
            {
            return_value = 1;
            break;
            }
          }

        i++;
        }

      fclose(tmp_file);
      }
    }
  else
    syslog(LOG_ERR, SYSLOG_ERROR_OPEN_KEYWORDS "%s", keywords_filename);

  return(return_value);
  }

int check_earlytalker(int delay)
  {
  int return_value;
  struct timeval tmp_time;
  fd_set listen_set;

  return_value = 0;

  tmp_time.tv_sec = delay;
  tmp_time.tv_usec = 0;

  FD_ZERO(&listen_set);
  FD_SET(STDIN_FD, &listen_set);
  if (select(STDIN_FD + 1, &listen_set, NULL, NULL, &tmp_time) == 1)
    return_value = 1;

  return(return_value);
  }

int dns_txt(char *target_name, struct filter_settings *current_settings)
  {
  static int initialized = 0;
  int return_value;
  unsigned char answer[PACKETSZ];
  unsigned char host[128];
  unsigned char *answer_ptr;
  int answer_length;
  int exp;
  int cttl;
  int size;
  int txt_length;
  int type;

  return_value = 0;

  if (!initialized)
    {
    if (res_init() >= 0)
      initialized = 1;
    else
      syslog(LOG_ERR, SYSLOG_ERROR_RES_INIT);
    }

  if (initialized)
    {
    memset(answer, 0, PACKETSZ);

    if ((answer_length = res_query(target_name, C_IN, T_TXT, answer, PACKETSZ)) >= 0)
      {
      answer_ptr = answer + sizeof(HEADER);
      if ((exp = dn_expand(answer, answer + answer_length, answer_ptr, (char *)host, sizeof(host))) >= 0)
        {
        answer_ptr += exp;
        GETSHORT(type, answer_ptr);
        if (type == T_TXT)
          {
          answer_ptr += INT16SZ; /* class */
          if ((exp = dn_expand(answer, answer + answer_length, answer_ptr, (char *)host, sizeof(host))) >= 0)
            {
            answer_ptr += exp;
            GETSHORT(type, answer_ptr);
            if (type == T_TXT)
              {
              answer_ptr += INT16SZ; /* class */
              GETLONG(cttl, answer_ptr);
              GETSHORT(size, answer_ptr);
              txt_length = *answer_ptr;
              if ((txt_length >= 0) &&
                  (txt_length < size))
                {
                answer_ptr++;
                sprintf(current_settings->reject_message_buf, "%.*s", MINVAL(txt_length, MAX_BUF - 2), answer_ptr);

                current_settings->reject_message = current_settings->reject_message_buf;
                current_settings->strlen_reject_message = MINVAL(txt_length, MAX_BUF - 2) + 2;
                current_settings->short_reject_message = SHORT_ERROR_RBL;

                return_value = 1;
                }
              else
                syslog(LOG_ERR, SYSLOG_ERROR_DNS_RESPONSE);
              }
            else
              syslog(LOG_ERR, SYSLOG_ERROR_TYPE_TXT);
            }
          else
            syslog(LOG_ERR, SYSLOG_ERROR_DN_EXPAND);
          }
        else
          syslog(LOG_ERR, SYSLOG_ERROR_TYPE_TXT);
        }
      else
        syslog(LOG_ERR, SYSLOG_ERROR_DN_EXPAND);
      }
    }

  return(return_value);
  }

int exec_path(char *filename, char *argv[], char *envp[])
  {
  char *path;
  char *tmp_start;
  char *tmp_end;
  char new_filename[MAX_BUF + 1];

  execve(filename, argv, envp);

  if (strchr(filename, '/') == NULL)
    {
    if (((path = find_environment_variable(envp, ENVIRONMENT_PATH, STRLEN_ENVIRONMENT_PATH)) == NULL) ||
        (path[0] == '\0'))
      path = DEFAULT_PATH;

    tmp_start = path;
    tmp_end = NULL;
    while (tmp_start != NULL)
      {
      if ((tmp_end = strchr(tmp_start, ':')) != NULL)
        {
        snprintf(new_filename, MAX_BUF, "%.*s/%s", tmp_end - tmp_start, tmp_start, filename);
        tmp_start = tmp_end + 1;
        }
      else
        {
        snprintf(new_filename, MAX_BUF, "%s/%s", tmp_start, filename);
        tmp_start = NULL;
        }

      execve(new_filename, argv, envp);
      }
    }

  return(0);
  }

int main(int argc, char *argv[], char *envp[])
  {
  struct filter_settings tmp_settings;
  int opt;
  int output_pipe[2];
  int input_pipe[2];
  int child_pid;
  int do_filter;
  int ip_in_name;
  int tmp_int;
  char rbl_name[MAX_BUF + 1];
  char ip_octets[4][4];

  openlog(SYSLOG_IDENTIFIER, LOG_PID, LOG_MAIL);

  init_settings(&tmp_settings);
  do_filter = -1;
  ip_in_name = -1;

  tmp_settings.server_name = find_environment_variable(envp, ENVIRONMENT_REMOTE_NAME, STRLEN_ENVIRONMENT_REMOTE_NAME);
  tmp_settings.strlen_server_name = (tmp_settings.server_name != NULL) ? strlen(tmp_settings.server_name) : 0;
  tmp_settings.server_ip = find_environment_variable(envp, ENVIRONMENT_REMOTE_IP, STRLEN_ENVIRONMENT_REMOTE_IP);
  tmp_settings.strlen_server_ip = (tmp_settings.server_ip != NULL) ? strlen(tmp_settings.server_ip) : 0;

  while ((opt = getopt(argc, argv, "clrRa:b:B:d:e:g:k:L:m:M:s:t:T:w:W:u:x:")) != -1)
    {
    switch (opt)
      {
      case 'a':
        if ((do_filter != 0) &&
            (sscanf(optarg, "%d", &tmp_settings.max_rcpt_to) == 1) &&
            (tmp_settings.max_rcpt_to > 0))
          do_filter = 1;

        break;
      case 'b':
        if ((do_filter != 0) &&
            (tmp_settings.reject_message == NULL) &&
            (search_file(optarg, tmp_settings.server_name, tmp_settings.strlen_server_name, '\0') == 1))
          {
          tmp_settings.reject_message = ERROR_BLACKLIST_NAME;
          tmp_settings.strlen_reject_message = STRLEN_ERROR_BLACKLIST_NAME;
          tmp_settings.short_reject_message = SHORT_ERROR_BLACKLIST_NAME;

          do_filter = 1;
          }

        break;
      case 'B':
        if ((do_filter != 0) &&
            (tmp_settings.reject_message == NULL) &&
            (search_file(optarg, tmp_settings.server_ip, tmp_settings.strlen_server_ip, '\0') == 1))
          {
          tmp_settings.reject_message = ERROR_BLACKLIST_IP;
          tmp_settings.strlen_reject_message = STRLEN_ERROR_BLACKLIST_IP;
          tmp_settings.short_reject_message = SHORT_ERROR_BLACKLIST_IP;

          do_filter = 1;
          }

        break;
      case 'c':
        if ((do_filter != 0) &&
            (tmp_settings.reject_message == NULL))
          {
          if (ip_in_name == -1)
            ip_in_name = is_ip_in_name(tmp_settings.server_ip, tmp_settings.server_name);

          if (ip_in_name &&
              check_country_code(tmp_settings.server_name, tmp_settings.strlen_server_name))
            {
            tmp_settings.reject_message = ERROR_IP_IN_NAME_CC;
            tmp_settings.strlen_reject_message = STRLEN_ERROR_IP_IN_NAME_CC;
            tmp_settings.short_reject_message = SHORT_ERROR_IP_IN_NAME_CC;

            do_filter = 1;
            }
          }

        break;
      case 'd':
        tmp_settings.local_domains_path = optarg;
        break;
      case 'e':
        if ((do_filter != 0) &&
            (tmp_settings.reject_message == NULL) &&
            ((tmp_int = atoi(optarg)) > 0) &&
            check_earlytalker(tmp_int))
          {
          tmp_settings.reject_message = ERROR_EARLYTALKER;
          tmp_settings.strlen_reject_message = STRLEN_ERROR_EARLYTALKER;
          tmp_settings.short_reject_message = SHORT_ERROR_EARLYTALKER;

          do_filter = 1;
          }

        break;
      case 'g':
        if ((do_filter != 0) &&
            (tmp_settings.reject_message == NULL))
          {
          tmp_settings.graylist_dir = optarg;
          do_filter = 1;
          }

        break;
      case 'k':
        if ((do_filter != 0) &&
            (tmp_settings.reject_message == NULL))
          {
          if (ip_in_name == -1)
            ip_in_name = is_ip_in_name(tmp_settings.server_ip, tmp_settings.server_name);

          if (ip_in_name &&
              check_rdns_keywords(optarg, tmp_settings.server_name, tmp_settings.strlen_server_name))
            {
            tmp_settings.reject_message = ERROR_IP_IN_NAME;
            tmp_settings.strlen_reject_message = STRLEN_ERROR_IP_IN_NAME;
            tmp_settings.short_reject_message = SHORT_ERROR_IP_IN_NAME;

            do_filter = 1;
            }
          }

        break;
      case 'l':
        tmp_settings.use_syslog = 1;
        if (do_filter != 0)
          do_filter = 1;

        break;
      case 'L':
        tmp_settings.log_dir = optarg;
        if (do_filter != 0)
          do_filter = 1;

        break;
      case 'm':
        sscanf(optarg, "%ld", &tmp_settings.graylist_min_secs);

        break;
      case 'M':
        sscanf(optarg, "%ld", &tmp_settings.graylist_max_secs);

        break;
      case 'r':
        if ((do_filter != 0) &&
            (tmp_settings.reject_message == NULL) &&
            ((tmp_settings.server_name == NULL) ||
             (tmp_settings.server_name[0] == '\0')))
          {
          tmp_settings.reject_message = ERROR_RDNS_MISSING;
          tmp_settings.strlen_reject_message = STRLEN_ERROR_RDNS_MISSING;
          tmp_settings.short_reject_message = SHORT_ERROR_RDNS_MISSING;

          do_filter = 1;
          }

        break;
      case 'R':
        if ((do_filter != 0) &&
            (tmp_settings.reject_message == NULL) &&
            (((strncmp(tmp_settings.server_name, LOCALHOST_NAME, STRLEN_LOCALHOST_NAME) != 0) &&
              !check_rdns_resolve(tmp_settings.server_name)) ||
             ((strncmp(tmp_settings.server_name, LOCALHOST_NAME, STRLEN_LOCALHOST_NAME) == 0) &&
              (strncmp(tmp_settings.server_ip, LOCALHOST_IP, STRLEN_LOCALHOST_IP) != 0))))
          {
          tmp_settings.reject_message = ERROR_RDNS_RESOLVE;
          tmp_settings.strlen_reject_message = STRLEN_ERROR_RDNS_RESOLVE;
          tmp_settings.short_reject_message = SHORT_ERROR_RDNS_RESOLVE;

          do_filter = 1;
          }

        break;
      case 's':
        if ((do_filter != 0) &&
            (tmp_settings.reject_message == NULL) &&
            (optarg != NULL))
          {
          tmp_settings.blacklist_sender_path = optarg;

          do_filter = 1;
          }

        break;
      case 't':
        if ((do_filter != 0) &&
            (sscanf(optarg, "%d", &tmp_settings.timeout_connection) == 1) &&
            (tmp_settings.timeout_connection > 0))
          do_filter = 1;

        break;
      case 'T':
        if ((do_filter != 0) &&
            (sscanf(optarg, "%d", &tmp_settings.timeout_command) == 1) &&
            (tmp_settings.timeout_command > 0))
          do_filter = 1;

        break;
      case 'w':
        if ((do_filter != 0) &&
            (search_file(optarg, tmp_settings.server_name, tmp_settings.strlen_server_name, '\0') == 1))
          do_filter = 0;

        break;
      case 'W':
        if ((do_filter != 0) &&
            (search_file(optarg, tmp_settings.server_ip, tmp_settings.strlen_server_ip, '\0') == 1))
          do_filter = 0;

        break;
      case 'x':
        if ((do_filter != 0) &&
            (tmp_settings.reject_message == NULL) &&
            (sscanf(tmp_settings.server_ip, "%3[0-9].%3[0-9].%3[0-9].%3[0-9]", ip_octets[0], ip_octets[1], ip_octets[2], ip_octets[3]) == 4) &&
            snprintf(rbl_name, MAX_BUF, "%s.%s.%s.%s.%s", ip_octets[3], ip_octets[2], ip_octets[1], ip_octets[0], optarg) &&
            dns_txt(rbl_name, &tmp_settings))
          do_filter = 1;

        break;
      case 'u':
        tmp_settings.policy_location = optarg;
        if ((tmp_settings.strlen_policy_location = strlen(tmp_settings.policy_location)) > MAX_POLICY)
          tmp_settings.strlen_policy_location = MAX_POLICY;

        break;
      default:
        usage();
      }
    }

  argv += optind;
  if (*argv == NULL)
    usage();

  output_pipe[0] = -1;
  output_pipe[1] = -1;
  input_pipe[0] = -1;
  input_pipe[1] = -1;

  if (find_environment_variable(envp, ENVIRONMENT_TLS, STRLEN_ENVIRONMENT_TLS))
    do_filter = 0;

  if (do_filter == 1)
    {
    if (tmp_settings.reject_message != NULL)
      {
      log_write(&tmp_settings, LOG_ACTION_FILTER_FROM, STDOUT_FD, REJECT_SEVERITY_HELO, STRLEN_REJECT_SEVERITY);
      log_write(&tmp_settings, LOG_ACTION_FILTER_FROM, STDOUT_FD, tmp_settings.reject_message, tmp_settings.strlen_reject_message);
      if (tmp_settings.policy_location != NULL)
        {
        log_write(&tmp_settings, LOG_ACTION_FILTER_FROM, STDOUT_FD, ERROR_URL, STRLEN_ERROR_URL);
        log_write(&tmp_settings, LOG_ACTION_FILTER_FROM, STDOUT_FD, tmp_settings.policy_location, tmp_settings.strlen_policy_location);
        }
      log_write(&tmp_settings, LOG_ACTION_FILTER_FROM, STDOUT_FD, REJECT_CRLF, STRLEN_REJECT_CRLF);

      middleman(&input_pipe[0], &output_pipe[1], '\n', &tmp_settings);
      }
    else if (pipe(output_pipe) != -1)
      {
      if (pipe(input_pipe) != -1)
        {
        if ((child_pid = fork()) > 0)
          // parent
          {
          close(output_pipe[0]);
          close(input_pipe[1]);

          middleman(&input_pipe[0], &output_pipe[1], '\n', &tmp_settings);

          if (output_pipe[1] != -1)
            close(output_pipe[1]);
          if (input_pipe[0] != -1)
            close(input_pipe[0]);
          wait(NULL);
          }
        else if (child_pid == 0)
          // child
          {
          close(output_pipe[1]);
          close(input_pipe[0]);

          if ((dup2(output_pipe[0], STDIN_FD) != -1) &&
              (dup2(input_pipe[1], STDOUT_FD) != -1))
            {
            if (!exec_path(*argv, argv, envp))
              {
              syslog(LOG_ERR, SYSLOG_ERROR_EXEC "%s: %s", *argv, strerror(errno));
              close(output_pipe[0]);
              close(input_pipe[1]);
              }
            }
          else
            {
            close(output_pipe[0]);
            close(input_pipe[1]);

            syslog(LOG_CRIT, SYSLOG_ERROR_MOVE_DESCRIPTORS "%s", strerror(errno));
            do_filter = 0;
            }
          }
        else
          {
          close(output_pipe[0]);
          close(output_pipe[1]);
          close(input_pipe[0]);
          close(input_pipe[1]);

          syslog(LOG_CRIT, SYSLOG_ERROR_FORK "%s", strerror(errno));
          do_filter = 0;
          }
        }
      else
        {
        close(output_pipe[0]);
        close(output_pipe[1]);

        syslog(LOG_CRIT, SYSLOG_ERROR_PIPE "%s", strerror(errno));
        do_filter = 0;
        }
      }
    else
      {
      syslog(LOG_CRIT, SYSLOG_ERROR_PIPE "%s", strerror(errno));
      do_filter = 0;
      }
    }

  if ((do_filter != 1) &&
      !exec_path(*argv, argv, envp))
    syslog(LOG_ERR, SYSLOG_ERROR_EXEC "%s: %s", *argv, strerror(errno));

  log_write(NULL, 0, -1, NULL, 0);
  closelog();

  return(0);
  }
